#pragma inline

/*

CA.C  Copyright (C) 1988  Mark Adler  Pasadena, CA
      All rights reserved.

Version history -

1.0      5 Jun 1988     First public version
1.1     18 Nov 1988     Fixed for Turbo C 2.0 (label needs statement)
1.2      3 Feb 1989     Display files and directories in lower case
1.3     21 Oct 1989     Catch dangling options and ignore command

CA is a command used to change and examine file attributes.  Entering
just:

     ca

may give a display like:

     *.*
       rhy- ibmbio.com
       rhy- ibmdos.com
       r--- command.com
       ---m config.sys
       ---m autoexec.bat

which is a list of files in the root directory and their attributes.
The first column of attributes is either '-' or 'r' for read/write or
read/only.  The second column is either '-' or 'h' for visible or
hidden.  The third column is '-' or 'y' for typical or system, and the
fourth column is '-' or 'm' for backed-up or modified.

To change attributes, there are eight options to set each of the four
attribute bits to their possible values.  For example, the command:

     ca/b

would try to change all the files in the current directory to the
backed-up status and would display, for the example directory above:

      changed config.sys
      changed autoexec.bat
     2 files' attributes changed

Note that CA only changed the files that needed to be changed.  See
below for a complete listing of the options.

CA can take file names to narrow down what is changed.  For example:

     ca/vt *.com

will make all the .COM files visible and typical, so for the above
directory, the system files IBMBIO.COM and IBMDOS.COM would show up in
the normal directory listing.

CA can also perform these operations to all the files contained in a
subdirectory and its subsubdirectories, etc.  The option for this is
"/S":

     ca/sr \dos

will make all the files in \DOS read-only.

Alternatively, CA can change all the files that match a particular name,
but in all the subdirectories.  The option for this is "/F" (find):

     ca/fr \*.exe

will make all the .EXE files on the current drive read/only, no matter
where they are.

CA can prompt you before changing the attribute of each file.  For
example:

     ca/pb *.*

will, for each file in the current directory that has been modified
since the last backup, ask youif it should change the modified bit or
not:

     Change file foo.c (y/n)? _

You must then answer y, n, Y, or N.  Any other response is ignored.

CA takes options that are preceded by a slash (/).  The command line is
read from left to right, processing options and file/path names as they
appear.  However, options that are appended to a file/path name take
effect before that name is processed.  For example, this command would
do the same thing as "ca/r *.exe":

     ca *.exe/r

But the command "ca *.exe /r" (notice the extra space before the /r)
would not do the same thing.  In this case, CA notices the "dangling"
option before it does anything and simply ignores the entire command for
safety, printing:

    Dangling option---entire command ignored

Since there are options to change the defaults along a command line,
there are options to change them back for processing the next thing
on the command line.  All the options are listed here:

     /R  - Change attribute to read-only.
     /W  - Change attribute to read-write.
     /H  - Change attribute to hidden.
     /V  - Change attribute to visible.
     /Y  - Change attribute to system.
     /T  - Change attribute to typical.
     /M  - Change attribute to modified.
     /B  - Change attribute to backed-up.
     /L  - No change---just list files and their attributes (default).
     /S  - Include matching subdirectories and all of their contents.
     /N  - Ignore subdirectories (default).
     /F  - Include all files that match pattern in all subdirs.
     /A  - Include all files in subdirectories, period (default).
     /P  - Prompt for each deletion or removal.
     /Q  - Quiet---don't prompt (default).
     /?  - Display version number and list of options.

A complicated example of CA is:

     ca /frp \*.com \*.exe /l \*.com \*.exe

which prompts to change all the .COM and .EXE files on the current drive
to read-only and then lists all the .COM and .EXE files on the current
drive along with their (possibly new) attributes.  The /L turns off the
change turned on by the /R.  The /P is ignored when listing, so it was
not necessary to turn that off with /Q.

*/


typedef unsigned long ulong;


/* Option flags, assigned to default values */
char att = 0,           /* Attribute for read-only, hidden, system */
     msk = 0,           /* Mask for attributes to change */
     sub = 0,           /* Include subdirectories */
     fin = 0,           /* Find---use same name in subdirs */
     ask = 0;           /* Prompt for each deletion */


ulong fls;              /* Number of files changed */


/* Structure for fnd1st(), fndnxt() */
struct find {
  char rsvd[21];        /* What DOS needs to keep track of find */
  char attr;            /* File attribute */
  unsigned time;        /* Time stamp */
  unsigned date;        /* Date stamp */
  ulong size;           /* File size in bytes */
  char name[13];        /* FIle name as a zero terminated string */
};


/* Send character 'c' to stdout */
void pputc(char c)
{
  asm mov DL,c
  asm mov AH,2
  asm int 21h
}


/* Send string 's' to stdout */
void pputs(char *s)
{
  asm mov SI,s
  asm cld
    plp:
  asm  lodsb
  asm  test AL,AL
  asm  jz pfin
  asm  mov DL,AL
  asm  mov AH,2
  asm  int 21h
  asm  jmp short plp
    pfin: ;
}


/* Convert the string s from upper case to lower case */
char *strlow(char *s)
{
  asm mov AX,DS
  asm mov ES,AX
  asm cld
  asm mov SI,s
  asm lea DI,[SI-1]
    llp:
  asm  inc DI
    llpnoi:
  asm  lodsb
  asm  test AL,AL
  asm  jz lfin
  asm  cmp AL,'A'
  asm  jb llp
  asm  cmp AL,'Z'
  asm  ja llp
  asm   add AL,'a'-'A'
  asm   stosb
  asm   jmp short llpnoi
    lfin: ;
  return s;
}


/* Find first match to 'p', attribute 'a', results in 'd' */
int fnd1st(char *p, struct find *f, int a)
{
  /* Set DTA to f */
  asm mov DX,f
  asm mov AH,1Ah
  asm int 21h

  /* Do find first */
  asm mov DX,p
  asm mov CX,a
  asm mov AH,4Eh
  asm int 21h
  asm sbb AX,AX
  return _AX;
}


/* Find next match for fnd1st() done on 'd' */
int fndnxt(struct find *f)
{
  /* Set DTA to f */
  asm mov DX,f
  asm mov AH,1Ah
  asm int 21h

  /* Do find next */
  asm mov AH,4Fh
  asm int 21h
  asm sbb AX,AX
  return _AX;
}


/* Copy string s to string d, return end of d */
char *scpy(char *d, char *s)
{
  asm mov SI,s
  asm mov DI,d
  asm cld
  asm mov AX,DS
  asm mov ES,AX
    slp:
  asm  lodsb
  asm  stosb
  asm  test AL,AL
  asm  jnz slp
  asm mov AX,DI
  asm dec AX
  return (char *) _AX;
}


/* Convert unsigned long 'n' to decimal in string 's' */
char *ultod(ulong n, char *s)
{
  /* Load n into BX:AX, s into SI and DI, and the radix into CX */
  asm mov DI,s
  asm mov SI,DI
  asm mov AX,n
  asm mov BX,n+2
  asm mov CX,10

  /* Convert n into a digit string, least significant digit first */
     dlp:
        /* Divide BX:AX by CX, quotient to BX:AX, remainder to DX */
  asm  xchg AX,BX       /* BX = low n */
  asm  sub DX,DX        /* DX:AX = high n */
  asm  div CX           /* AX = high q, DX = temporary r */
  asm  xchg AX,BX       /* BX = high q, DX:AX = temp r:low q */
  asm  div CX           /* BX:AX = q, DX = r */
        /* Put digit in string */
  asm  add DL,'0'
  asm  mov [DI],DL
  asm  inc DI
        /* Do until BX:AX is zero */
  asm  mov DX,AX
  asm  or DX,BX
  asm  jnz dlp
        /* Terminate string */
  asm mov [DI],AL

  /* Reverse the string, putting most significant digit first */
     rlp:
  asm  dec DI
  asm  cmp DI,SI
  asm  jna rfin
  asm  mov AL,[SI]
  asm  xchg AL,[DI]
  asm  mov [SI],AL
  asm  inc SI
  asm  jmp short rlp
     rfin:

  /* Return pointer to start of string */
  asm mov AX,s
  return (char *) _AX;
}


/* Ask user yes/no question */
int query(char *s, char *t)
{
  pputs(s);
  pputs(t);
  pputs(" (y/n)? ");

     qlp:
  asm  mov AH,8         /* Get console input without echo */
  asm  int 21h
  asm  test AL,AL
  asm  jnz got
  asm   int 21h         /* If extended code, ignore it */
  asm   jmp short qlp
      got:
  asm  mov DL,AL        /* Save character typed */
  asm  and AL,5Fh       /* Convert lower to upper case */
  asm  sub SI,SI        /* Return 0 if no */
  asm  cmp AL,'N'
  asm  je qfin
  asm  inc SI           /* Return 1 if yes */
  asm  cmp AL,'Y'
  asm  jne qlp          /* Do until we get a proper answer */
     qfin:

  pputc(_DX);           /* Echo y, n, Y, or N */
  pputs("\r\n");
  return _SI;
}


void change(char *a)
{
  register char *p, *q;
  int e, f;
  struct find d;
  char s[128];
  char t[128];

  /* Setup */
  q = s;
  p = scpy(q, a) - 1;           /* Copy path being searched */
  if (p < s || *p == '\\' || *p == ':')
    scpy(p + 1, "*.*");         /* If no name, append wildcard */
  else
    do {
      p--;                      /* Scan back for path delimiter */
    } while (p >= s && *p != '\\' && *p != ':');
  p++;                          /* Point past delimiter */
  if (sub || fin)
    scpy(t, p);                 /* Save name to match */
  f = 1;                        /* Haven't printed dir name yet */

  /* Find matching files */
  e = fnd1st(q, &d, 7);         /* Search all attributes */
  while (!e)
  {
    if (msk)
    {                           /* Change file attributes */
      if ((d.attr & msk) != att)        /* See if change needed */
      {
        scpy(p, d.name);
        if (!ask || query("Change file ", strlow(q)))
        {
          _CX = (d.attr & ~msk) | att;    /* New attribute */
          asm mov DX,q
          asm mov AX,4301h
          asm int 21h
          asm jnb delok
            pputs("Could not change attributes of file ");
            pputs(strlow(q));
            pputs("\r\n");
            asm jmp short delbad
         delok:
            pputs(" changed ");
            pputs(strlow(q));
            pputs("\r\n");
            fls++;
         delbad: ;
        }
      }
    }
    else                        /* Show file attributes */
    {
      if (f)                    /* Show directory name if needed */
      {
        pputs(strlow(q));
        pputs("\r\n");
        f = 0;
      }
      pputs("  ");
      pputc(d.attr & 1 ? 'r' : '-');    /* Show attributes */
      pputc(d.attr & 2 ? 'h' : '-');
      pputc(d.attr & 4 ? 'y' : '-');
      pputc(d.attr & 32 ? 'm' : '-');
      pputc(' ');
      pputs(strlow(d.name));    /* Show name */
      pputs("\r\n");
    }
    e = fndnxt(&d);
  }

  /* Find matching subdirectories, if requested */
  if (sub || fin)
  {
    scpy(p, fin ? "*.*" : t);   /* What directories to match */
    e = fnd1st(q, &d, 0x17);    /* Look at everything except labels */
    while (!e)                  /* Do all subdirectories */
    {
      if ((d.attr & 0x10) &&
          (d.name[0] != '.' || (d.name[1] && d.name[1] != '.')))
      {                         /* Is subdir and not "." or ".." */
        q = scpy(p, d.name);    /* Overlay wildcard with found name */
        *q++ = '\\';            /* Append path delimiter */
        scpy(q, fin ? t : "*.*");       /* New wildcard */
        change(s);              /* Do subdirectory */
      }
      e = fndnxt(&d);
    }
  }
}


void main(int argc, char *argv[])
{
  register char *p;
  register int k;
  int i, m, c, n;
  char *a[64];          /* Tokens and options */
  char t[64];           /* Token/~option flags */


  /* Parse line into tokens and options */
  m = k = 0;
  for (i = 1; i < argc; i++)    /* Do command line */
  {
    p = argv[i];                /* Next argument */
    if (*p != '/')              /* See if token */
    {
      a[k] = p++;
      t[k++] = 1;               /* Token */
      m = 1;                    /* Token flag */
      while (*p && *p != '/')
        p++;
      if (*p)                   /* See if options on token */
      {
        *p++ = '\0';            /* Terminate token string */
        a[k] = a[k-1];          /* Put options appended to token */
        t[k] = 1;               /*  BEFORE that token. */
        a[k-1] = p;
        t[k-1] = 0;             /* Option(s) */
        k++;
      }
    }
    else
    {
      a[k] = ++p;               /* Options start after the slash */
      t[k++] = 0;               /* Option(s) */
      if (m)
        m = -1;                 /* Dangling option flag */
    }
  }
  if (m < 0)
  {
    pputs("Dangling option---entire command ignored\r\n");
    asm mov AX,4C01h            /* Exit with error */
    asm int 21h
  }
  m = k;                        /* Number of tokens and options */

  /* Process tokens and options */
  fls = 0;                      /* Initialize files changed */
  c = 0;                        /* Change requested */
  n = 0;                        /* change() called at all */
  for (i = 0; i < m; i++)
    if (t[i])                   /* Token */
    {
      c |= msk;                 /* Update changes requested flag */
      n |= 1;                   /* Update change() called flag */
      change(a[i]);             /* Do it */
    }
    else                        /* Options */
      for (p = a[i]; (k = *p) != 0; p++)
        if (k == '/')       {}          /* Ignore extra /'s */
        else if ((k &= 0x5f) == 'Q')  ask = 0;  /* Don't prompt */
        else if (k == 'P')  ask = 1;    /* Prompt */
        else if (k == 'W') {att &= ~1;  msk |= 1; } /* Read-write */
        else if (k == 'R') {att |= 1;   msk |= 1; } /* Read-only */
        else if (k == 'V') {att &= ~2;  msk |= 2; } /* Visible */
        else if (k == 'H') {att |= 2;   msk |= 2; } /* Hidden */
        else if (k == 'T') {att &= ~4;  msk |= 4; } /* Typical */
        else if (k == 'Y') {att |= 4;   msk |= 4; } /* System */
        else if (k == 'B') {att &= ~32; msk |= 32;} /* Backed up */
        else if (k == 'M') {att |= 32;  msk |= 32;} /* Modified */
        else if (k == 'L')  att = msk = 0;      /* Just list */
        else if (k == 'N')  sub = 0;    /* No subdirs */
        else if (k == 'S')  sub = 1;    /* Subdirs */
        else if (k == 'A')  fin = 0;    /* Use *.* in subs */
        else if (k == 'F')  fin = 1;    /* Find file */
        else                            /* Invalid option */
        {
          if (*p != '?')
            pputs("Invalid option\r\n");
          pputs("\
CA 1.3  Copyright (C) 1988,1989  Mark Adler  All rights reserved.\r\n\
Valid options are (*=default):\r\n\
 /R\t- Change to read-only\r\n\
 /W\t- Change to read-write\r\n\
 /H\t- Change to hidden\r\n\
 /V\t- Change to visible\r\n\
 /Y\t- Change to system\r\n\
 /T\t- Change to typical\r\n\
 /M\t- Change to modified\r\n\
 /B\t- Change to backed-up\r\n\
 /L\t- No change---just list *\r\n\
 /S\t- Include subdirectories\r\n\
 /N\t- Ignore subdirectories *\r\n\
 /F\t- Find files\r\n\
 /A\t- All files *\r\n\
 /P\t- Prompt for each change\r\n\
 /Q\t- No prompting *\r\n\
 /?\t- Display this list\r\n");
            asm mov AX,4C01h    /* Exit with error */
            asm int 21h
        }
  if (!n)                       /* If change() never called, */
  {                             /*  then call it */
    c |= msk;                   /* Update changes requested flag */
    change("");
  }
  if (c)                        /* If change was requested, */
  {                             /*  show how many were changed */
    pputs(ultod(fls, t));
    pputs(" file");
    if (fls != 1)
      pputs("s'");
    else
      pputs("'s");
    pputs(" attributes changed\r\n");
  }
}
